unit GameUnit;

{$mode objfpc}{$H+}

interface

uses Classes, SysUtils, LCLType, nxTypes, nxGLExtra;

const
  SectorSize = 100000.0;
  WorldSize = 30;
  ParallaxCount = 2000;
  ParallaxDistance = 100.0;
  PlanetTypes = 10;
  MaxBeams = 2000;

type
  TShipType = record
    name: string;
  end;

  TSecLoc = record
    x, y: smallint;
  end;

  TShip = class
    position, movement, warpOrigin, warpTarget: TVector2f;
    rotX, rotZ, armor, warpDist, warpProgress: single;
    target: TShip;
    secLoc, warpLoc: TSecLoc;
    owner, sType: shortint;
    warping: byte;
    inFleet, accelerating: boolean;
    fireDelay: word;
  end;

  TCamera = record
    position, movement: TVector2f;
    rotX, rotY, toRotX, toRotY, zoom, toZoom: single;
    secLoc: TSecLoc;
  end;

  TSun = record
    color: TRGB;
  end;
  pSun = ^TSun;

  TPlanetType = record
    texture: integer;
    name: string;
  end;
  pPlanetType = ^TPlanetType;

  TPlanet = record
    size, pop: single;
    position: TVector2f;
    owner, pType: shortint;
  end;
  pPlanet = ^TPlanet;

  TSector = record
    name: string;
    sun: pSun;
    ships: integer;
    planet: array of TPlanet;
    planets, back: shortint;
    known: boolean;
  end;
  pSector = ^TSector;

  TPlayer = record
    ship: array of TShip;
    ships: integer;
    warpSec: TSecLoc;
    warpPos: TVector2f;
    inCombat: boolean;
  end;

  TBeam = record
    p1, p2: TVector2f;
    color: TRGBA;
  end;

  TParallax = record
    position: TVector;
    color: TRGB;
  end;

  { TGame }

  TGame = class
  private
  public
    cam: TCamera;
    pl: array[0..7] of TPlayer;
    sector: array[0..WorldSize-1, 0..WorldSize-1] of TSector;
    parallax: array[0..ParallaxCount-1] of TParallax;
    pType: array[0..PlanetTypes-1] of TPlanetType;
    beam: array[0..MaxBeams-1] of TBeam;
    beamCount: integer;
    pt: TParticleEngine;
    constructor Create;
    procedure AddLog(s: string);
    function AddShip(plIndex: shortint; sec: TSecLoc; _pos: TVector2f): TShip;
    procedure ChangeCamSector(addX, addY: integer);
    procedure Clear;
    function CurSector: pSector;
    function GetPos(const p: TVector2f; const secLoc: TSecLoc): TVector2f;
    procedure GameLoop;
    procedure InitiateWarp(s: TShip; sec: TSecLoc; _pos: TVector2f);
    function IsEnemyPlanets(loc: TSecLoc; plIndex: integer): boolean;
    function IsPlayerSector(loc: TSecLoc): boolean;
    function IsEnemiesInSector(loc: TSecLoc; plIndex: integer): boolean;
    function MaxArmor(_stype: integer): single;
    function MaxPop(planet: pPlanet): single;
    function NearPlanet(s: TShip): pPlanet;
    function NearestEnemy(s: TShip): TShip;
    procedure NewGame;
    function RelativePos(sourcePos, destPos: TVector2f;
      sourceLoc, destLoc: TSecLoc): TVector2f;
    procedure RemoveShip(ship: TShip);
    procedure RemoveShip(plIndex, sIndex: integer);
    procedure UpdateShipSec(ship: TShip);
    destructor Destroy; override;
  end;

var
  game: TGame;
  plShip: TShip;
  keys: array[0..255] of boolean;
  curPos: TVector;
  systemView: boolean;
  focX, focY: integer;
  textLog: TStringList;
  pTypeStr: array[0..9] of string = (
    'Gas giant',
    'Radiated',
    'Toxic',
    'Barren',
    'Desert',
    'Tundra',
    'Ocean',
    'Swamp',
    'Terran',
    'Gaia');

  function GetSecLoc(x, y: smallint): TSecLoc;
  function Match(const a, b: TSecLoc): boolean;

implementation

uses math, nxMath, nxMath3D;

function Match(const a, b: TSecLoc): boolean;
begin
  result:=(a.x=b.x) and (a.y=b.y);
end;

function GetSecLoc(x, y: smallint): TSecLoc;
begin
  result.x:=x; result.y:=y;
end;

{ TGame }

constructor TGame.Create;
begin
  pType[0].name:='Gas Giant';
  pType[1].name:='Radiated';
  pType[2].name:='Toxic';
  pType[3].name:='Barren';
  pType[4].name:='Desert';
  pType[5].name:='Tundra';
  pType[6].name:='Ocean';
  pType[7].name:='Swamp';
  pType[8].name:='Terran';
  pType[9].name:='Gaia';
  NewGame;
end;

procedure TGame.AddLog(s: string);
begin
  while textLog.Count<5 do textLog.Add('');
  textLog.Add(s);
  while textLog.Count>5 do textLog.Delete(0);
end;

procedure TGame.NewGame;
var x, y, i, j, n, plIndex: integer;
    a, d: single; sl: TStringList; tmp: string;
    pS: TShip;
begin
  Clear;

  // Add stars
  sl:=TStringList.Create;
  sl.LoadFromFile('data\StarNames.txt');
  for i:=0 to sl.Count-1 do begin
    n:=random(sl.Count);
    if (sl[i]<>'') and (sl[n]<>'') then begin
      tmp:=sl[i]; sl[i]:=sl[n]; sl[n]:=tmp;
    end;
  end;
  plIndex:=0;
  for j:=min(sl.Count-1, 20+random(10)) downto 0 do begin
    if plIndex=0 then begin
      x:=WorldSize div 5+random(WorldSize div 3+1);
      y:=WorldSize div 5+random(WorldSize div 3+1);
    end else begin
      repeat
        x:=random(worldSize); y:=random(worldSize);
      until sector[x, y].sun=nil;
    end;
    with sector[x, y] do begin
      name:=sl[j];
      sun:=getmem(sizeof(TSun));
      sun^.color:=rgb(200+random(56), 200+random(56), 200+random(56));
      n:=random(7);
      case plIndex of
        1, 2: back:=plIndex;
        else back:=0;
      end;
      planets:=random(6)*random(2);
      if plIndex>-1 then planets:=max(planets, 1);
      setlength(planet, planets);
      n:=random(planets);
      for i:=0 to planets-1 do begin
        planet[i].size:=10+random*5;
        a:=random*2*pi; d:=(i+1)*(SectorSize*0.05)+random*(SectorSize*0.01);
        planet[i].position:=vector2f(SectorSize*0.5+sin(a)*d, SectorSize*0.5+cos(a)*d);
        if (i=n) and (plIndex>-1) then begin
          // Colonized planet
          if plIndex=0 then planet[i].pType:=8
          else planet[i].pType:=8+plIndex mod 2;
          planet[i].owner:=plIndex;
          planet[i].pop:=10;
          // Set player ship on home planet orbit
          if plIndex=0 then begin
            known:=true;
            // Mothership for player
            a:=random*2*pi; d:=planet[i].size*1.1;
            plShip:=AddShip(plIndex, GetSecLoc(x, y), vector2f(
              planet[i].position.x+cos(a)*d,
              planet[i].position.y+sin(a)*d));
            plShip.sType:=0; plShip.inFleet:=true;
            // Combat ships for player
            AddShip(plIndex, GetSecLoc(x, y), planet[i].position);
            // Add ship for player fleet
            pS:=AddShip(plIndex, GetSecLoc(x, y), planet[i].position);
            pS.inFleet:=true;
          end else begin
            // Mothership for AI
            pS:=AddShip(plIndex, GetSecLoc(x, y), planet[i].position);
            pS.sType:=0; pS.inFleet:=true;
            // Combat ship for AI
            AddShip(plIndex, GetSecLoc(x, y), planet[i].position);
          end;
          plIndex:=plIndex+1;
          if plIndex>high(pl) then plIndex:=-1;
        end else begin
          planet[i].pType:=random(PlanetTypes);
          planet[i].owner:=-1;
          planet[i].pop:=0;
        end;
      end;
    end;
    repeat
      x:=random(WorldSize); y:=random(WorldSize);
    until (sector[x, y].sun=nil) or (j=0);
  end;

  sl.Free;
  if plShip<>nil then begin
    cam.secLoc:=plShip.secLoc;
    cam.movement:=plShip.movement;
    cam.position:=plShip.position;
  end;
  cam.zoom:=10; cam.toZoom:=cam.zoom;
  cam.rotX:=plShip.rotX; cam.rotY:=30;
  cam.toRotX:=cam.rotX; cam.toRotY:=cam.rotY;
  systemView:=true;
  focX:=cam.secLoc.x; focY:=cam.secLoc.y;
  for i:=0 to ParallaxCount-1 do begin
    parallax[i].position:=vector(0+(random*2-1)*ParallaxDistance,
      (random*2-1)*ParallaxDistance, 0+(random*2-1)*ParallaxDistance);
    n:=random(200)+56;
    parallax[i].color:=rgb(n-random(10), n-random(10), n-random(10));
  end;
end;

function TGame.AddShip(plIndex: shortint; sec: TSecLoc; _pos: TVector2f): TShip;
begin
  if (plIndex<0) or (plIndex>high(pl)) then begin
    nxSetError('AddShip: plIndex out of bounds'); exit;
  end;
  inc(pl[plIndex].ships); setlength(pl[plIndex].ship, pl[plIndex].ships);
  pl[plIndex].ship[pl[plIndex].ships-1]:=TShip.Create;
  result:=pl[plIndex].ship[pl[plIndex].ships-1];
  result.secLoc:=sec;
  result.sType:=1;
  result.armor:=MaxArmor(result.sType);
  result.rotX:=random*360;
  result.rotZ:=0;
  result.position:=_pos;
  result.movement:=vector2f(0, 0);
  result.inFleet:=false;
  result.target:=nil;
  result.warping:=0;
  result.fireDelay:=60;
  inc(sector[sec.x, sec.y].ships);
end;

procedure TGame.ChangeCamSector(addX, addY: integer);
begin
  cam.secLoc.x:=cam.secLoc.x+addX;
  cam.secLoc.y:=cam.secLoc.y+addY;
end;

procedure TGame.Clear;
var i, j, x, y: integer;
begin
  for y:=0 to WorldSize-1 do
    for x:=0 to WorldSize-1 do
      with sector[x, y] do begin
        if sun<>nil then begin
          FreeMem(sun); sun:=nil;
        end;
        known:=false; planets:=0; ships:=0; name:='';
      end;
  for i:=0 to high(pl) do begin
    if pl[i].ships>0 then
      for j:=0 to pl[i].ships-1 do pl[i].ship[j].Free;
    pl[i].ships:=0; pl[i].warpSec.x:=-1; pl[i].warpSec.y:=-1;
  end;
  beamCount:=0;
end;

function TGame.CurSector: pSector;
begin
  result:=@sector[cam.secLoc.x, cam.secLoc.y];
end;

function TGame.GetPos(const p: TVector2f; const secLoc: TSecLoc): TVector2f;
begin
  result:=vector2f(p.x+secLoc.x*SectorSize, p.y+secLoc.y*SectorSize);
end;

procedure TGame.GameLoop;
var i, j, k: integer; d, a: single; p, tgt, moveAmount: TVector2f;
    newship: TShip; nearP: pPlanet; ppt: pParticle;
begin
  moveAmount:=vector2f(0,0);

  // Beams
  for i:=beamCount-1 downto 0 do
    with beam[i] do begin
      if color.a<10 then begin
        beam[i]:=beam[beamCount-1]; dec(beamCount);
      end else
        color.a:=color.a-10;
    end;

  // Player controls
  if plShip<>nil then
    with plShip do begin
      if warping=0 then begin
        accelerating:=keys[ord('W')];
        if accelerating then begin
          d:=angle(rotX*toRad, (cam.rotX-90)*toRad)*toDeg;
          if d<-90 then d:=-90
          else if d>90 then d:=90;
          rotX:=rotX-d*0.03;
          if abs(d)<20 then begin
            movement.x:=interpolate(movement.x, cos(rotX*toRad)*0.8, 0.02);
            movement.y:=interpolate(movement.y, sin(rotX*toRad)*0.8, 0.02);
          end;
        end else if keys[ord('S')] then begin
          movement.x:=interpolate(movement.x, 0, 0.03);
          movement.y:=interpolate(movement.y, 0, 0.03);
        end;
        if keys[ord('A')] then begin
          movement.x:=interpolate(movement.x, cos((rotX-90)*toRad)*0.16, 0.02);
          movement.y:=interpolate(movement.y, sin((rotX-90)*toRad)*0.16, 0.02);
        end;
        if keys[ord('D')] then begin
          movement.x:=interpolate(movement.x, cos((rotX+90)*toRad)*0.16, 0.02);
          movement.y:=interpolate(movement.y, sin((rotX+90)*toRad)*0.16, 0.02);
        end;
      end;
    end; // with plShip

  // Colonies
  for j:=0 to worldSize-1 do
    for i:=0 to worldSize-1 do
      with sector[i, j] do
        for k:=0 to planets-1 do
          with planet[k] do begin
            if (owner>=0) and (random<pop*0.000005) then begin
              // Create ships randomly
              newship:=AddShip(owner, getSecLoc(i, j), position);
              newship.owner:=owner;
              // Half defend, half join the fleet
              newship.inFleet:=random<0.5;
              if pl[owner].ships>1 then
                if newship.inFleet then begin
                  InitiateWarp(newship, pl[owner].ship[0].secloc,
                    pl[owner].ship[0].position);
                  game.AddLog('New ship warping to fleet');
                end else
                  game.AddLog('New ship built for sector defence');
            end;
            if pop<MaxPop(@planet[k]) then
              pop:=pop+0.001;
          end;

  // Ships
  for i:=0 to high(pl) do
    with pl[i] do begin
      for j:=ships-1 downto 0 do
        with ship[j] do begin
          if fireDelay>0 then dec(fireDelay);

          // Ship warping
          if warping>0 then begin
            accelerating:=true;
            p:=RelativePos(position, warpTarget, secLoc, warpLoc);
            d:=angle(0,0, p.x,p.y);
            d:=angle(rotX*toRad, d-0)*toDeg;
            if d<-90 then d:=-90
            else if d>90 then d:=90;
            rotX:=rotX-d*0.05;
            if warping<255 then begin
              inc(warping);
              if warping=255 then begin
                movement:=vector2f(0, 0);
                warpOrigin:=GetPos(position, secloc);
                p:=GetPos(position, warpLoc);
                warpDist:=hypot(warpOrigin.x-p.x, warpOrigin.y-p.y);
              end else if abs(d)<20 then begin
                // Move slowly towards warptarget
                movement.x:=interpolate(movement.x, cos(rotX*toRad)*0.8, 0.02);
                movement.y:=interpolate(movement.y, sin(rotX*toRad)*0.8, 0.02);
              end;
            end else begin
              // Warp movement
              movement.x:=0; movement.y:=0;
              warpProgress:=warpProgress+500/warpDist;
              if warpProgress>1 then begin
                // Arrive to warp destination
                warpProgress:=1; warping:=0;
                if owner=0 then sector[warpLoc.x, warpLoc.y].known:=true;
                warpSec.x:=-1; accelerating:=false;
                movement.x:=cos(rotX*toRad)*0.1;
                movement.y:=sin(rotX*toRad)*0.1;
              end;
              d:=smoothen(warpProgress);
              p:=GetPos(warpTarget, warpLoc);
              position.x:=interpolate(warpOrigin.x, p.x, d);
              position.y:=interpolate(warpOrigin.y, p.y, d);
              secloc.x:=trunc(position.x/sectorSize);
              secloc.y:=trunc(position.y/sectorSize);
              position.x:=position.x-secLoc.x*sectorSize;
              position.y:=position.y-secLoc.y*sectorSize;
            end;

          end else if stype=1 then begin
            // Combat ship
            accelerating:=false;
            if target=nil then begin
              if inFleet then begin
                tgt:=vector2f(
                  ship[0].position.x+cos((360*j/ships)*toRad)*5,
                  ship[0].position.y+sin((360*j/ships)*toRad)*5);
                if (ship[0].secLoc.x<>secloc.x) or (ship[0].secLoc.y<>secloc.y) then
                  InitiateWarp(ship[j], ship[0].secLoc, ship[0].position);
              end else begin
                if target=nil then begin
                  nearP:=NearPlanet(ship[j]);
                  tgt:=vector2f(
                    nearP^.position.x+cos((360*j/ships)*toRad)*20,
                    nearP^.position.y+sin((360*j/ships)*toRad)*20);
                  if random<0.001 then target:=NearestEnemy(ship[j]);
                end;
              end;
            end else begin
              tgt:=target.position;
            end;
            d:=angle(position.x,position.y, tgt.x,tgt.y);
            d:=angle(rotX*toRad, d)*toDeg;
            if d<-90 then d:=-90
            else if d>90 then d:=90;
            rotX:=rotX-d*0.05;
            if abs(d)<20 then accelerating:=true;

            if accelerating then begin
              movement.x:=interpolate(movement.x, cos(rotX*toRad)*0.8, 0.03);
              movement.y:=interpolate(movement.y, sin(rotX*toRad)*0.8, 0.03);
            end;

          end else if (stype=0) and (owner<>0) then begin
            // Enemy mothership
            if IsEnemiesInSector(secloc, i) then begin
              if inCombat then begin
                // Mothership AI combat
                if target=nil then target:=NearestEnemy(ship[j]);
              end else if (not inCombat) and (random<0.01) then
                inCombat:=true;
            end else inCombat:=false;
            if not inCombat then begin
              // Mothership idle AI
              movement.x:=interpolate(movement.x, cos(rotX*toRad)*0.08, 0.03);
              movement.y:=interpolate(movement.y, sin(rotX*toRad)*0.08, 0.03);
              if random<0.001 then begin
                // Warp fleet to random sector
                repeat
                  warpSec.x:=random(worldSize);
                  warpSec.y:=random(worldSize);
                until sector[warpSec.x, warpSec.y].planets>0;
                warpPos:=sector[warpSec.x, warpSec.y].planet[
                  random(sector[warpSec.x, warpSec.y].planets)].position;
                warpPos.x:=warpPos.x+random*100-50;
                warpPos.y:=warpPos.y+random*100-50;
                for k:=0 to ships-1 do
                  if ship[k].inFleet then
                    InitiateWarp(ship[k], warpSec, warpPos);
              end;
            end;
          end else if stype=0 then begin
            // Player's mothership
            if inCombat then begin
              if target=nil then begin
                target:=NearestEnemy(ship[j]);
                if target=nil then begin
                  // Capture planet
                  nearP:=NearPlanet(ship[j]);
                  if (nearP<>nil) and (nearP^.owner<>owner) then begin
                    nearP^.owner:=owner;
                    if nearP^.pop<1 then nearP^.pop:=1;
                    game.AddLog('Planet captured');
                  end;
                  inCombat:=false;
                  game.AddLog('Combat ended');
                end;
              end;
            end;
          end;

          // Combat actions
          if (target<>nil) and ((target.secLoc.x<>secLoc.x) or (target.secLoc.y<>secLoc.y)) then
            target:=nil;
          if (target<>nil) and (warping=0) then begin
            if ship[j]<>plShip then begin
              // Move towards target
              d:=angle(position.x,position.y, target.position.x,target.position.y);
              d:=angle(rotX*toRad, d)*toDeg;
              if d<-90 then d:=-90
              else if d>90 then d:=90;
              rotX:=rotX-d*0.05;
              if abs(d)<20 then begin
                accelerating:=true;
                movement.x:=interpolate(movement.x, cos(rotX*toRad)*0.8, 0.03);
                movement.y:=interpolate(movement.y, sin(rotX*toRad)*0.8, 0.03);
              end;
            end;
            // Ship in firing range
            if pointInCircle(position, target.position, 100) then begin
              if fireDelay=0 then begin
                //if owner=0 then game.AddLog('*DEBUG: Firing*');
                fireDelay:=50;
                if (cam.secLoc.x=secLoc.x) and (cam.secLoc.y=secLoc.y) then begin
                  if beamCount<MaxBeams then inc(beamCount);
                  with beam[beamCount-1] do begin
                    color:=rgba(255,0,0,255);
                    p1:=position;
                    p2:=target.position;
                  end;
                end;
                target.armor:=target.armor-3-random*2;
                if target.armor<=0 then begin
                  // Ship explosion
                  if (cam.secLoc.x=secLoc.x) and (cam.secLoc.y=secLoc.y) then begin
                    for k:=1 to 500 do begin
                      a:=random*2*pi;
                      d:=random*0.2;
                      ppt:=pt.AddParticle(0, 0.5+0.5*random, 0.5+0.5*random, 0.2,
                        0.3, target.position.x, 0, target.position.y);
                      ppt^.velocity.x:=target.movement.x+cos(a)*d;
                      ppt^.velocity.y:=(random*1-0.5)*d;
                      ppt^.velocity.z:=target.movement.y+sin(a)*d;
                    end;
                  end;
                  RemoveShip(target);
                end;
              end;
            end;

          end;

          if warping<255 then position:=VectorAdd(position, movement);
          if ship[j]=plShip then begin
            moveAmount.x:=-movement.x;
            moveAmount.y:=-movement.y;
          end;
          UpdateShipSec(ship[j]);
          if rotX<0 then rotX:=rotX+360
          else if rotX>=360 then rotX:=rotX-360;

        end;
    end;

  // Camera
  while cam.rotX<cam.toRotX-180 do
    cam.rotX:=cam.rotX+360;
  while cam.rotX>cam.toRotX+180 do
    cam.rotX:=cam.rotX-360;
  cam.rotX:=interpolate(cam.rotX, cam.toRotX, 0.1);
  cam.rotY:=interpolate(cam.rotY, cam.toRotY, 0.1);
  cam.zoom:=interpolate(cam.zoom, cam.toZoom, 0.1);

  if plShip<>nil then begin
    {if (cam.secLoc.x<>plShip.secLoc.x) or (cam.secLoc.y<>plShip.secLoc.y) then begin
      game.ChangeCamSector(plShip.secLoc.x-cam.secLoc.x,
        plShip.secLoc.y-cam.secLoc.y);
    end; }
    cam.secLoc:=plShip.secLoc;
    cam.position:=plShip.position;
  end;

  // Update parallax
  for i:=0 to ParallaxCount-1 do
    with parallax[i].position do begin
      x:=x+moveAmount.x; z:=z+moveAmount.y;
      if x>ParallaxDistance then x:=x-ParallaxDistance*2
      else if x<-ParallaxDistance then x:=x+ParallaxDistance*2;
      if z>ParallaxDistance then z:=z-ParallaxDistance*2
      else if z<-ParallaxDistance then z:=z+ParallaxDistance*2;
    end;
end;

procedure TGame.InitiateWarp(s: TShip; sec: TSecLoc; _pos: TVector2f);
var p: TVector2f; a, d: single;
begin
  if s.warping>0 then exit;
  s.warping:=1;
  s.warpLoc:=sec;
  s.warpOrigin:=GetPos(s.position, s.secloc);
  a:=random*2*pi;
  d:=random*30;
  s.warpTarget.x:=_pos.x+cos(a)*d;
  s.warpTarget.y:=_pos.y+sin(a)*d;
  p:=GetPos(_pos, sec);
  s.warpDist:=hypot(s.warpOrigin.x-p.x, s.warpOrigin.y-p.y);
  s.warpProgress:=0;
  s.target:=nil;
  if s.sType=0 then pl[s.owner].inCombat:=false;
end;

function TGame.IsEnemyPlanets(loc: TSecLoc; plIndex: integer): boolean;
var i: integer;
begin
  result:=false;
  with sector[loc.x, loc.y] do
    for i:=0 to planets-1 do
      if (planet[i].owner<>plIndex) and (planet[i].owner>=0) then result:=true;
end;

function TGame.IsPlayerSector(loc: TSecLoc): boolean;
var i: integer;
begin
  result:=false;
  with sector[loc.x, loc.y] do
    for i:=0 to planets-1 do
      if planet[i].owner=0 then result:=true;
end;

function TGame.IsEnemiesInSector(loc: TSecLoc; plIndex: integer): boolean;
var i, j: integer;
begin
  result:=false;
  for j:=0 to high(pl) do
    if j<>plIndex then
      for i:=0 to pl[j].ships-1 do
        with pl[j].ship[i] do
          if (secloc.x=loc.x) and (secloc.y=loc.y) then begin
            result:=true; exit;
          end;
end;

function TGame.MaxArmor(_stype: integer): single;
begin
  case _stype of
    0: result:=400;
    else result:=100;
  end;
end;

function TGame.MaxPop(planet: pPlanet): single;
begin
  if planet^.pType<3 then result:=0
  else result:=(planet^.size)*(planet^.pType);
end;

function TGame.NearPlanet(s: TShip): pPlanet;
var i: integer; d, nearest: single;
begin
  result:=nil;
  if s=nil then exit;
  with sector[s.secLoc.x, s.secLoc.y] do begin
    nearest:=SectorSize;
    for i:=0 to planets-1 do begin
      d:=hypot(s.position.x-planet[i].position.x, s.position.y-planet[i].position.y);
      if d<nearest then begin
        nearest:=d; result:=@planet[i];
      end;
    end;
  end;
end;

function TGame.NearestEnemy(s: TShip): TShip;
var i, j: integer; d, nearest: single;
begin
  result:=nil; nearest:=SectorSize;
  for j:=0 to high(pl) do
    if j<>s.owner then
      for i:=0 to pl[j].ships-1 do
        with pl[j].ship[i] do
          if (secloc.x=s.secloc.x) and (secloc.y=s.secloc.y) then begin
            d:=hypot(position.x-s.position.x, position.y-s.position.y);
            if d<nearest then begin
              nearest:=d; result:=pl[j].ship[i];
            end;
          end;
end;

function TGame.RelativePos(sourcePos, destPos: TVector2f; sourceLoc, destLoc: TSecLoc): TVector2f;
begin
  result.x:=destLoc.x*SectorSize+destPos.x-sourceLoc.x*SectorSize-sourcePos.x;
  result.y:=destLoc.y*SectorSize+destPos.y-sourceLoc.y*SectorSize-sourcePos.y;
end;

procedure TGame.RemoveShip(ship: TShip);
var i: integer;
begin
  for i:=0 to pl[ship.owner].ships-1 do
    if pl[ship.owner].ship[i]=ship then begin
      RemoveShip(ship.owner, i); exit;
    end;
end;

procedure TGame.RemoveShip(plIndex, sIndex: integer);
var s: TShip; i, j: integer;
begin
  s:=pl[plIndex].ship[sIndex];
  for j:=0 to high(pl) do
    for i:=0 to pl[j].ships-1 do
      if pl[j].ship[i].target=s then pl[j].ship[i].target:=nil;
  if s=plShip then begin
    plShip:=nil;
  end;
  if s.sType=0 then pl[plIndex].warpSec.x:=-1;
  s.Free;
  with pl[plIndex] do begin
    ship[sIndex]:=ship[ships-1]; dec(ships);
  end;
end;

procedure TGame.UpdateShipSec(ship: TShip);
  procedure ChangeSec(dx, dy: integer);
  begin
    with ship do begin
      dec(sector[secloc.x, secloc.y].ships);
      secLoc.x:=secLoc.x+dx;
      secLoc.y:=secLoc.y+dy;
      inc(sector[secloc.x, secloc.y].ships);
    end;
  end;
begin
  if ship=nil then exit;
  with ship do begin
    if (position.x<0) and (secLoc.x>0) then begin
      position.x:=position.x+SectorSize;
      ChangeSec(-1, 0);
    end else if (position.x>=SectorSize) and (secLoc.x<WorldSize-1) then begin
      position.x:=position.x-SectorSize;
      ChangeSec(1, 0);
    end;
    if (position.y<0) and (secLoc.y>0) then begin
      position.y:=position.y+SectorSize;
      ChangeSec(0, -1);
    end else if (position.y>=SectorSize) and (secLoc.y<WorldSize-1) then begin
      position.y:=position.y-SectorSize;
      ChangeSec(0, 1);
    end;
  end;
end;

destructor TGame.Destroy;
begin
  Clear;
  inherited Destroy;
end;

end.

